@@ -1,16 +1,20 @@
 import sys
-from typing import Any, MutableMapping
+from typing import Any, Callable, MutableMapping
 
 import anyio
 import pytest
 from anyio.abc import ObjectReceiveStream, ObjectSendStream
 
 from starlette import status
-from starlette.types import Receive, Scope, Send
+from starlette.responses import Response
+from starlette.testclient import TestClient, WebSocketDenialResponse
+from starlette.types import Message, Receive, Scope, Send
 from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState
 
+TestClientFactory = Callable[..., TestClient]
 
-def test_websocket_url(test_client_factory):
+
+def test_websocket_url(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -23,7 +27,7 @@ def test_websocket_url(test_client_factory):
         assert data == {"url": "ws://testserver/123?a=abc"}
 
 
-def test_websocket_binary_json(test_client_factory):
+def test_websocket_binary_json(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -38,7 +42,25 @@ def test_websocket_binary_json(test_client_factory):
         assert data == {"test": "data"}
 
 
-def test_websocket_query_params(test_client_factory):
+def test_websocket_ensure_unicode_on_send_json(
+    test_client_factory: TestClientFactory,
+) -> None:
+    async def app(scope: Scope, receive: Receive, send: Send) -> None:
+        websocket = WebSocket(scope, receive=receive, send=send)
+
+        await websocket.accept()
+        message = await websocket.receive_json(mode="text")
+        await websocket.send_json(message, mode="text")
+        await websocket.close()
+
+    client = test_client_factory(app)
+    with client.websocket_connect("/123?a=abc") as websocket:
+        websocket.send_json({"test": "数据"}, mode="text")
+        data = websocket.receive_text()
+        assert data == '{"test":"数据"}'
+
+
+def test_websocket_query_params(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         query_params = dict(websocket.query_params)
@@ -56,7 +78,7 @@ def test_websocket_query_params(test_client_factory):
     any(module in sys.modules for module in ("brotli", "brotlicffi")),
     reason='urllib3 includes "br" to the "accept-encoding" headers.',
 )
-def test_websocket_headers(test_client_factory):
+def test_websocket_headers(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         headers = dict(websocket.headers)
@@ -79,7 +101,7 @@ def test_websocket_headers(test_client_factory):
         assert data == {"headers": expected_headers}
 
 
-def test_websocket_port(test_client_factory):
+def test_websocket_port(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -92,7 +114,9 @@ def test_websocket_port(test_client_factory):
         assert data == {"port": 123}
 
 
-def test_websocket_send_and_receive_text(test_client_factory):
+def test_websocket_send_and_receive_text(
+    test_client_factory: TestClientFactory,
+) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -107,7 +131,9 @@ def test_websocket_send_and_receive_text(test_client_factory):
         assert data == "Message was: Hello, world!"
 
 
-def test_websocket_send_and_receive_bytes(test_client_factory):
+def test_websocket_send_and_receive_bytes(
+    test_client_factory: TestClientFactory,
+) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -122,7 +148,9 @@ def test_websocket_send_and_receive_bytes(test_client_factory):
         assert data == b"Message was: Hello, world!"
 
 
-def test_websocket_send_and_receive_json(test_client_factory):
+def test_websocket_send_and_receive_json(
+    test_client_factory: TestClientFactory,
+) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -137,7 +165,7 @@ def test_websocket_send_and_receive_json(test_client_factory):
         assert data == {"message": {"hello": "world"}}
 
 
-def test_websocket_iter_text(test_client_factory):
+def test_websocket_iter_text(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -151,7 +179,7 @@ def test_websocket_iter_text(test_client_factory):
         assert data == "Message was: Hello, world!"
 
 
-def test_websocket_iter_bytes(test_client_factory):
+def test_websocket_iter_bytes(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -165,7 +193,7 @@ def test_websocket_iter_bytes(test_client_factory):
         assert data == b"Message was: Hello, world!"
 
 
-def test_websocket_iter_json(test_client_factory):
+def test_websocket_iter_json(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -179,17 +207,17 @@ def test_websocket_iter_json(test_client_factory):
         assert data == {"message": {"hello": "world"}}
 
 
-def test_websocket_concurrency_pattern(test_client_factory):
+def test_websocket_concurrency_pattern(test_client_factory: TestClientFactory) -> None:
     stream_send: ObjectSendStream[MutableMapping[str, Any]]
     stream_receive: ObjectReceiveStream[MutableMapping[str, Any]]
     stream_send, stream_receive = anyio.create_memory_object_stream()
 
-    async def reader(websocket):
+    async def reader(websocket: WebSocket) -> None:
         async with stream_send:
             async for data in websocket.iter_json():
                 await stream_send.send(data)
 
-    async def writer(websocket):
+    async def writer(websocket: WebSocket) -> None:
         async with stream_receive:
             async for message in stream_receive:
                 await websocket.send_json(message)
@@ -209,25 +237,50 @@ def test_websocket_concurrency_pattern(test_client_factory):
         assert data == {"hello": "world"}
 
 
-def test_client_close(test_client_factory):
+def test_client_close(test_client_factory: TestClientFactory) -> None:
     close_code = None
+    close_reason = None
 
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
-        nonlocal close_code
+        nonlocal close_code, close_reason
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
         try:
             await websocket.receive_text()
         except WebSocketDisconnect as exc:
             close_code = exc.code
+            close_reason = exc.reason
 
     client = test_client_factory(app)
     with client.websocket_connect("/") as websocket:
-        websocket.close(code=status.WS_1001_GOING_AWAY)
+        websocket.close(code=status.WS_1001_GOING_AWAY, reason="Going Away")
     assert close_code == status.WS_1001_GOING_AWAY
+    assert close_reason == "Going Away"
 
 
-def test_application_close(test_client_factory):
+@pytest.mark.anyio
+async def test_client_disconnect_on_send() -> None:
+    async def app(scope: Scope, receive: Receive, send: Send) -> None:
+        websocket = WebSocket(scope, receive=receive, send=send)
+        await websocket.accept()
+        await websocket.send_text("Hello, world!")
+
+    async def receive() -> Message:
+        return {"type": "websocket.connect"}
+
+    async def send(message: Message) -> None:
+        if message["type"] == "websocket.accept":
+            return
+        # Simulate the exception the server would send to the application when the
+        # client disconnects.
+        raise OSError
+
+    with pytest.raises(WebSocketDisconnect) as ctx:
+        await app({"type": "websocket", "path": "/"}, receive, send)
+    assert ctx.value.code == status.WS_1006_ABNORMAL_CLOSURE
+
+
+def test_application_close(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -240,19 +293,126 @@ def test_application_close(test_client_factory):
         assert exc.value.code == status.WS_1001_GOING_AWAY
 
 
-def test_rejected_connection(test_client_factory):
+def test_rejected_connection(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
+        msg = await websocket.receive()
+        assert msg == {"type": "websocket.connect"}
         await websocket.close(status.WS_1001_GOING_AWAY)
 
     client = test_client_factory(app)
     with pytest.raises(WebSocketDisconnect) as exc:
         with client.websocket_connect("/"):
-            pass  # pragma: nocover
+            pass  # pragma: no cover
     assert exc.value.code == status.WS_1001_GOING_AWAY
 
 
-def test_subprotocol(test_client_factory):
+def test_send_denial_response(test_client_factory: TestClientFactory) -> None:
+    async def app(scope: Scope, receive: Receive, send: Send) -> None:
+        websocket = WebSocket(scope, receive=receive, send=send)
+        msg = await websocket.receive()
+        assert msg == {"type": "websocket.connect"}
+        response = Response(status_code=404, content="foo")
+        await websocket.send_denial_response(response)
+
+    client = test_client_factory(app)
+    with pytest.raises(WebSocketDenialResponse) as exc:
+        with client.websocket_connect("/"):
+            pass  # pragma: no cover
+    assert exc.value.status_code == 404
+    assert exc.value.content == b"foo"
+
+
+def test_send_response_multi(test_client_factory: TestClientFactory) -> None:
+    async def app(scope: Scope, receive: Receive, send: Send) -> None:
+        websocket = WebSocket(scope, receive=receive, send=send)
+        msg = await websocket.receive()
+        assert msg == {"type": "websocket.connect"}
+        await websocket.send(
+            {
+                "type": "websocket.http.response.start",
+                "status": 404,
+                "headers": [(b"content-type", b"text/plain"), (b"foo", b"bar")],
+            }
+        )
+        await websocket.send(
+            {
+                "type": "websocket.http.response.body",
+                "body": b"hard",
+                "more_body": True,
+            }
+        )
+        await websocket.send(
+            {
+                "type": "websocket.http.response.body",
+                "body": b"body",
+            }
+        )
+
+    client = test_client_factory(app)
+    with pytest.raises(WebSocketDenialResponse) as exc:
+        with client.websocket_connect("/"):
+            pass  # pragma: no cover
+    assert exc.value.status_code == 404
+    assert exc.value.content == b"hardbody"
+    assert exc.value.headers["foo"] == "bar"
+
+
+def test_send_response_unsupported(test_client_factory: TestClientFactory) -> None:
+    async def app(scope: Scope, receive: Receive, send: Send) -> None:
+        del scope["extensions"]["websocket.http.response"]
+        websocket = WebSocket(scope, receive=receive, send=send)
+        msg = await websocket.receive()
+        assert msg == {"type": "websocket.connect"}
+        response = Response(status_code=404, content="foo")
+        with pytest.raises(
+            RuntimeError,
+            match="The server doesn't support the Websocket Denial Response extension.",
+        ):
+            await websocket.send_denial_response(response)
+        await websocket.close()
+
+    client = test_client_factory(app)
+    with pytest.raises(WebSocketDisconnect) as exc:
+        with client.websocket_connect("/"):
+            pass  # pragma: no cover
+    assert exc.value.code == status.WS_1000_NORMAL_CLOSURE
+
+
+def test_send_response_duplicate_start(test_client_factory: TestClientFactory) -> None:
+    async def app(scope: Scope, receive: Receive, send: Send) -> None:
+        websocket = WebSocket(scope, receive=receive, send=send)
+        msg = await websocket.receive()
+        assert msg == {"type": "websocket.connect"}
+        response = Response(status_code=404, content="foo")
+        await websocket.send(
+            {
+                "type": "websocket.http.response.start",
+                "status": response.status_code,
+                "headers": response.raw_headers,
+            }
+        )
+        await websocket.send(
+            {
+                "type": "websocket.http.response.start",
+                "status": response.status_code,
+                "headers": response.raw_headers,
+            }
+        )
+
+    client = test_client_factory(app)
+    with pytest.raises(
+        RuntimeError,
+        match=(
+            'Expected ASGI message "websocket.http.response.body", but got '
+            "'websocket.http.response.start'"
+        ),
+    ):
+        with client.websocket_connect("/"):
+            pass  # pragma: no cover
+
+
+def test_subprotocol(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         assert websocket["subprotocols"] == ["soap", "wamp"]
@@ -264,7 +424,7 @@ def test_subprotocol(test_client_factory):
         assert websocket.accepted_subprotocol == "wamp"
 
 
-def test_additional_headers(test_client_factory):
+def test_additional_headers(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept(headers=[(b"additional", b"header")])
@@ -275,7 +435,7 @@ def test_additional_headers(test_client_factory):
         assert websocket.extra_headers == [(b"additional", b"header")]
 
 
-def test_no_additional_headers(test_client_factory):
+def test_no_additional_headers(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -286,17 +446,17 @@ def test_no_additional_headers(test_client_factory):
         assert websocket.extra_headers == []
 
 
-def test_websocket_exception(test_client_factory):
+def test_websocket_exception(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         assert False
 
     client = test_client_factory(app)
     with pytest.raises(AssertionError):
         with client.websocket_connect("/123?a=abc"):
-            pass  # pragma: nocover
+            pass  # pragma: no cover
 
 
-def test_duplicate_close(test_client_factory):
+def test_duplicate_close(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -306,10 +466,10 @@ def test_duplicate_close(test_client_factory):
     client = test_client_factory(app)
     with pytest.raises(RuntimeError):
         with client.websocket_connect("/"):
-            pass  # pragma: nocover
+            pass  # pragma: no cover
 
 
-def test_duplicate_disconnect(test_client_factory):
+def test_duplicate_disconnect(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -323,17 +483,17 @@ def test_duplicate_disconnect(test_client_factory):
             websocket.close()
 
 
-def test_websocket_scope_interface():
+def test_websocket_scope_interface() -> None:
     """
     A WebSocket can be instantiated with a scope, and presents a `Mapping`
     interface.
     """
 
-    async def mock_receive():
-        pass  # pragma: no cover
+    async def mock_receive() -> Message:  # type: ignore
+        ...  # pragma: no cover
 
-    async def mock_send(message):
-        pass  # pragma: no cover
+    async def mock_send(message: Message) -> None:
+        ...  # pragma: no cover
 
     websocket = WebSocket(
         {"type": "websocket", "path": "/abc/", "headers": []},
@@ -355,7 +515,7 @@ def test_websocket_scope_interface():
     assert {websocket} == {websocket}
 
 
-def test_websocket_close_reason(test_client_factory) -> None:
+def test_websocket_close_reason(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -369,7 +529,7 @@ def test_websocket_close_reason(test_client_factory) -> None:
         assert exc.value.reason == "Going Away"
 
 
-def test_send_json_invalid_mode(test_client_factory):
+def test_send_json_invalid_mode(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -378,10 +538,10 @@ def test_send_json_invalid_mode(test_client_factory):
     client = test_client_factory(app)
     with pytest.raises(RuntimeError):
         with client.websocket_connect("/"):
-            pass  # pragma: nocover
+            pass  # pragma: no cover
 
 
-def test_receive_json_invalid_mode(test_client_factory):
+def test_receive_json_invalid_mode(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -393,7 +553,7 @@ def test_receive_json_invalid_mode(test_client_factory):
             pass  # pragma: nocover
 
 
-def test_receive_text_before_accept(test_client_factory):
+def test_receive_text_before_accept(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.receive_text()
@@ -404,7 +564,7 @@ def test_receive_text_before_accept(test_client_factory):
             pass  # pragma: nocover
 
 
-def test_receive_bytes_before_accept(test_client_factory):
+def test_receive_bytes_before_accept(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.receive_bytes()
@@ -415,7 +575,7 @@ def test_receive_bytes_before_accept(test_client_factory):
             pass  # pragma: nocover
 
 
-def test_receive_json_before_accept(test_client_factory):
+def test_receive_json_before_accept(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.receive_json()
@@ -423,10 +583,10 @@ def test_receive_json_before_accept(test_client_factory):
     client = test_client_factory(app)
     with pytest.raises(RuntimeError):
         with client.websocket_connect("/"):
-            pass  # pragma: nocover
+            pass  # pragma: no cover
 
 
-def test_send_before_accept(test_client_factory):
+def test_send_before_accept(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.send({"type": "websocket.send"})
@@ -437,7 +597,7 @@ def test_send_before_accept(test_client_factory):
             pass  # pragma: nocover
 
 
-def test_send_wrong_message_type(test_client_factory):
+def test_send_wrong_message_type(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.send({"type": "websocket.accept"})
@@ -446,10 +606,10 @@ def test_send_wrong_message_type(test_client_factory):
     client = test_client_factory(app)
     with pytest.raises(RuntimeError):
         with client.websocket_connect("/"):
-            pass  # pragma: nocover
+            pass  # pragma: no cover
 
 
-def test_receive_before_accept(test_client_factory):
+def test_receive_before_accept(test_client_factory: TestClientFactory) -> None:
     async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
@@ -462,8 +622,8 @@ def test_receive_before_accept(test_client_factory):
             websocket.send({"type": "websocket.send"})
 
 
-def test_receive_wrong_message_type(test_client_factory):
-    async def app(scope, receive, send):
+def test_receive_wrong_message_type(test_client_factory: TestClientFactory) -> None:
+    async def app(scope: Scope, receive: Receive, send: Send) -> None:
         websocket = WebSocket(scope, receive=receive, send=send)
         await websocket.accept()
         await websocket.receive()
